@@ -27,7 +27,7 @@ License:
 
 """
 
-import abc, collections, csv, io, io_wrapper, json, jsonpath_ng.ext, logging, re, requests, requests_cache, shutil, six, sys, tempfile, time, xlrd
+import abc, collections, csv, io, io_wrapper, json, jsonpath_ng.ext, logging, re, requests, requests_cache, shutil, six, sys, tempfile, time, xlrd3 as xlrd
 
 import hxl, hxl.filters
 import zipfile
@@ -54,7 +54,7 @@ GOOGLE_SHEETS_URL = r'^https?://[^/]+google.com/.*[^0-9A-Za-z_-]([0-9A-Za-z_-]{4
 GOOGLE_FILE_URL = r'https?://drive.google.com/file/d/([0-9A-Za-z_-]+)/.*$'
 DROPBOX_URL = r'^https://www.dropbox.com/s/([0-9a-z]{15})/([^?]+)\?dl=[01]$'
 CKAN_URL = r'^(https?://[^/]+)/dataset/([^/]+)(?:/resource/([a-z0-9-]{36}))?$'
-HXL_PROXY_SAVED_URL = r'^(https?://[^/]*proxy.hxlstandard.org)/data/([a-zA-Z0-9_]{6}).*$'
+HXL_PROXY_SAVED_URL = r'^(https?://[^/]*proxy.hxlstandard.org)/data/([a-zA-Z0-9_]{6})[^?]*(\?.*)?$'
 HXL_PROXY_ARGS_URL = r'^(https?://[^/]*proxy.hxlstandard.org)/data.*\?(.+)$'
 KOBO_URL = r'^https://kobo.humanitarianresponse.info/#/forms/([A-Za-z0-9]{16,32})/'
 
@@ -420,9 +420,11 @@ def make_input(raw_source, allow_local=False, sheet_index=None, timeout=None, ve
             tmpfile = make_tempfile(input)
 
             try:
+                # Is the zip file really an XLSX file?
                 logger.debug('Trying input from an Excel file')
                 return XLSXInput(tmpfile, sheet_index=sheet_index)
-            except:
+            except xlrd.XLRDError:
+                # If not, see if it contains a CSV file
                 if match_sigs(sig, ZIP_SIGS): # more-restrictive
                     zf = zipfile.ZipFile(tmpfile, "r")
                     for name in zf.namelist():
@@ -466,8 +468,27 @@ def open_url_or_file(url_or_filename, allow_local=False, timeout=None, verify_ss
     result = re.search(r'\.([A-Za-z0-9]{1,5})$', url_or_filename)
     if result:
         file_ext = result.group(1).lower()
-    
-    if re.match(r'^(?:https?|s?ftp)://', url_or_filename.lower()):
+
+    result = re.match(r'^(?:https?|s?ftp)://([^/]+)', url_or_filename.lower())
+    if result:
+
+        # Check for possible exploits when allow_local is False
+        if not allow_local:
+
+            hostname = result.group(1).lower().strip()
+
+            # forbid dotted quads
+            if re.match(r'^[0-9.]+$', hostname):
+                raise HXLIOException("Security settings forbid accessing host via IP address {}", hostname)
+
+            # forbid localhost
+            if hostname == "localhost":
+                raise HXLIOException("Security settings forbid accessing localhost")
+                
+            # forbid localhost
+            if hostname.endswith(".localdomain"):
+                raise HXLIOException("Security settings forbid accessing hostnames ending in .localdomain: {}", hostname)
+                
         # It looks like a URL
         file_ext = os.path.splitext(urllib.parse.urlparse(url_or_filename).path)[1]
         try:
@@ -508,7 +529,7 @@ def open_url_or_file(url_or_filename, allow_local=False, timeout=None, verify_ss
 
     else:
         # Forbidden to try local (allow_local is False), so give up.
-        logger.critical('Tried to open local file %s with allow_local set to False', url_or_filename)
+        logger.critical('Security settings forbid accessing local files or non http(s)/ftp(s) URL schemes: %s', url_or_filename)
         raise HXLIOException(
             "Only http(s) and (s)ftp URLs allowed: {}".format(url_or_filename),
             url=url_or_filename
@@ -761,8 +782,17 @@ class CSVInput(AbstractInput):
         @param encoding: the character encoding to use
         @returns a csv.Dialect object or string.
         """
+
+        raw = input.peek(16384)
+
+        # Special case: there might be part of a multibyte Unicode character at the end
+        for i in range(0, 7):
+            try:
+                sample = raw[:-1].decode(encoding)
+            except Exception as e:
+                continue
+            break
         
-        sample = input.peek(16384).decode(encoding)
         lines = re.split(r'\r?\n', sample)
 
         # first, try for a hashtag row
@@ -978,7 +1008,7 @@ class XLSInput(AbstractInput):
         super().__init__()
         self.tmpfile = tmpfile # prevent garbage collection as long as this object exists
         self.is_repeatable = True
-        self._workbook = xlrd.open_workbook(filename=tmpfile.name)
+        self._workbook = xlrd.open_workbook(filename=tmpfile.name, on_demand=True, ragged_rows=True)
         if sheet_index is None:
             sheet_index = self._find_hxl_sheet_index()
         self._sheet = self._workbook.sheet_by_index(sheet_index)
@@ -1074,12 +1104,7 @@ class XLSXInput(AbstractInput):
         super().__init__()
         self.tmpfile = tmpfile # prevent garbage collection
         self.is_repeatable = True
-        try:
-            self._workbook = xlrd.open_workbook(filename=tmpfile.name)
-        except TypeError as e:
-            # xlrd throws a TypeError when it's trying to open a non-XLSX zip
-            # there are probably other exceptions we need to trap here
-            raise HXLIOException("Not an Excel workbook (possibly a zip archive)")
+        self._workbook = xlrd.open_workbook(filename=tmpfile.name, on_demand=True, ragged_rows=True)
         if sheet_index is None:
             sheet_index = self._find_hxl_sheet_index()
         self._sheet = self._workbook.sheet_by_index(sheet_index)
@@ -1471,6 +1496,7 @@ def _munge_url(url, verify_ssl=True, http_headers=None):
         response = requests.head(url)
         if response.is_redirect:
             url = response.headers['Location']
+            logger.info("Following Google Drive redirect to %s", url)
 
     # Is it a Kobo survey?
     result = re.match(KOBO_URL, url)
@@ -1485,30 +1511,39 @@ def _munge_url(url, verify_ssl=True, http_headers=None):
     # Is it a Google Drive *file*?
     result = re.match(GOOGLE_FILE_URL, url)
     if result:
-        return 'https://drive.google.com/uc?export=download&id={}'.format(result.group(1))
+        url = 'https://drive.google.com/uc?export=download&id={}'.format(result.group(1))
+        logger.info("Google Drive direct file download URL: %s", url)
 
     # Is it a Google *Sheet*?
     result = re.match(GOOGLE_SHEETS_URL, url)
     if result and not re.search(r'/pub', url):
         if result.group(2):
-            return 'https://docs.google.com/spreadsheets/d/{0}/export?format=csv&gid={1}'.format(result.group(1), result.group(2))
+            url = 'https://docs.google.com/spreadsheets/d/{0}/export?format=csv&gid={1}'.format(result.group(1), result.group(2))
         else:
-            return 'https://docs.google.com/spreadsheets/d/{0}/export?format=csv'.format(result.group(1))
+            url = 'https://docs.google.com/spreadsheets/d/{0}/export?format=csv'.format(result.group(1))
+        logger.info("Google Sheets CSV export URL: %s", url)
+        return url
 
     # Is it a Dropbox URL?
     result = re.match(DROPBOX_URL, url)
     if result:
-        return 'https://www.dropbox.com/s/{0}/{1}?dl=1'.format(result.group(1), result.group(2))
+        url = 'https://www.dropbox.com/s/{0}/{1}?dl=1'.format(result.group(1), result.group(2))
+        logger.info("Dropbox direct-download URL: %s", url)
+        return url
 
     # Is it a HXL Proxy saved recipe?
     result = re.match(HXL_PROXY_SAVED_URL, url)
     if result:
-        return '{0}/data/{1}.csv'.format(result.group(1), result.group(2))
+        url = '{0}/data/{1}.csv{2}'.format(result.group(1), result.group(2), result.group(3))
+        logger.info("HXL Proxy saved-recipe URL: %s", url)
+        return url
 
     # Is it a HXL Proxy args-based recipe?
     result = re.match(HXL_PROXY_ARGS_URL, url)
     if result:
-        return '{0}/data.csv?{1}'.format(result.group(1), result.group(2))
+        url = '{0}/data.csv?{1}'.format(result.group(1), result.group(2))
+        logger.info("HXL Proxy direct-download URL: %s", url)
+        return url
 
     # No changes
     return url
@@ -1540,6 +1575,7 @@ def _get_ckan_url(site_url, dataset_id, resource_id, verify_ssl, http_headers):
         ckan_api_result = requests.get(ckan_api_query, verify=verify_ssl, headers=http_headers).json()
         if ckan_api_result['success']:
             url = ckan_api_result['result']['url']
+            logger.info("Converted CKAN resource URL to %s", url)
         elif ckan_api_result['error']['__type'] == 'Authorization Error':
             raise HXLAuthorizationException(
                 "Not authorised to read CKAN resource (is the dataset public?): {}".format(
@@ -1561,6 +1597,7 @@ def _get_ckan_url(site_url, dataset_id, resource_id, verify_ssl, http_headers):
         ckan_api_result = requests.get(ckan_api_query, verify=verify_ssl, headers=http_headers).json()
         if ckan_api_result['success']:
             url = ckan_api_result['result']['resources'][0]['url']
+            logger.info("Converted CKAN package URL to %s", url)
         elif ckan_api_result['error']['__type'] == 'Authorization Error':
             raise HXLAuthorizationException(
                 "Not authorised to read CKAN dataset (is it public?): {}".format(
@@ -1625,7 +1662,7 @@ def _get_kobo_url(asset_id, url, verify_ssl, http_headers, max_export_age_second
 
         # if less than four hours, and has a URL, use it (and stop here)
         if export.get('result') and (age_in_seconds < max_export_age_seconds):
-            logger.info("Reusing existing Kobo export for %s", asset_id)
+            logger.info("Reusing existing Kobo export for %s: %s", asset_id, export['result'])
             return export['result']
 
     logger.info("Generating new Kobo export for %s", asset_id)
@@ -1674,6 +1711,7 @@ def _get_kobo_url(asset_id, url, verify_ssl, http_headers, max_export_age_second
         url = response.json().get("result")
 
         if url:
+            logger.info("Kobo export URL: %s", url)
             return url
 
         fail_counter += 1
